// file:   rot_enc.c
// author: bto
// brief:  read the input from a rotary encoder

#include <xc.h> // XC8

#include "defs.h"
#include "delay.h"
#include "rot_enc.h"
#include "pic18f4550-config.h" // configuration bits

#define LED LATDbits.LATD3 // onboard LED
#define BUT PORTBbits.RB4  // onboard button

#define ENC_BUT PORTBbits.RB5  // rotary encoder button
#define ENC_A PORTBbits.RB6 // rotary encoder pin A
#define ENC_B PORTBbits.RB7 // rotary encoder pin B

int val;
int val_temp;

void
toggle_led(void)
{
    if (val != 0) {
        val_temp = val; // store current value
	val = 0;
    }
    else {
        val = val_temp; // set current value
    }
}

int enc_pos;
int enc_prev_pos;
int enc_dir;

void
handle_encoder(void)
{
    enc_dir = (enc_prev_pos << 2) | enc_pos;
    
    if (enc_dir == 0b1011 || enc_dir == 0b1101 || enc_dir == 0b0100 || enc_dir == 0b0010) {
        // clockwise step
	val = val + 10;
        if (val > 1023) val = 1023;

    } else if (enc_dir == 0b0001 || enc_dir == 0b0111 || enc_dir == 0b1110 || enc_dir == 0b1000) {
        // counter-clockwise step
        val = val - 10;
        if (val < 0) val = 0;
    }
    
    enc_prev_pos = enc_pos;
}

void 
__interrupt(high_priority, auto_psv) isr(void)
{
    if (INTCONbits.TMR0IE && INTCONbits.TMR0IF) {
        // handle timer0 overflow interrupt
        // toggle LED
        if (LED == ON) {
            LED = OFF;
        }
        else {
            LED = ON;
        }
        TMR0H = 0x4c;
        TMR0L = 0x4b;
        INTCONbits.TMR0IF = 0; // clear the TMR0 flag
        return;
    }
    if (INTCONbits.RBIE && INTCONbits.RBIF) {
        // button on = 1
	if (ENC_BUT) {
            toggle_led();
        }
	// button on = 0
	if (!BUT) {
            toggle_led();
	}
        enc_pos = (ENC_A << 1) | ENC_B;
	if (enc_prev_pos != enc_pos) {
            handle_encoder();
        }
        
	INTCONbits.RBIF = 0; // clear the RB port change flag
        return;
    }
    return;
}

void 
setup(void)
{
    PORTB = 0; // initialize PORTB by clearing output data latches

    // reset control register    
    RCONbits.IPEN = 0; // enable priority interrupts (1 = on, 0 = off)

    // oscillator control register
    OSCCONbits.IDLEN = 1; // idle enable (1 = idle, 0 = sleep)
    OSCCONbits.IRCF  = 6; // internal osc freq select (111 = 8MHz, 110 = 4MHz, 0 = 31kHz)
    OSCCONbits.OSTS  = 0; // start-up time-out (1 = ready, 0 = not ready)
    OSCCONbits.IOFS  = 0; // INTOSC freq stable (1 = stable, 0 = unstable)
    OSCCONbits.SCS   = 0; // system clock select (00 = primary osc, 01 = Timer1, 1x = Internal osc)

    // interrupt control register
    INTCONbits.RBIF   = 0; // RB port change interrupt flag 
    INTCONbits.INT0IF = 0; // external interrupt flag
    INTCONbits.TMR0IF = 0; // timer0 interrupt flag
    INTCONbits.RBIE   = 1; // RB port interrupt enable
    INTCONbits.INT0IE = 0; // external interrupt enable
    INTCONbits.TMR0IE = 0; // timer0 interrupt enable
    
    if (RCONbits.IPEN) {
        INTCONbits.GIEL = 1; // low priority interrupt enable (1 = enable, 0 = disable)
        INTCONbits.GIEH = 1; // high priority interrupt enable (1 = enable, 0 = disable)
    } else {
        INTCONbits.PEIE = 0; // peripheral interrupt enable (1 = enable, 0 = disable)
        INTCONbits.GIE  = 1; // global interrupt enable (1 = enable, 0 = disable)
    }

    // interrupt control 2 register
    INTCON2bits.RBPU    = 1; // Weak internal pull-up for PORTB pins (0 = enable, 1 = disable)
    INTCON2bits.INTEDG0 = 0; // Edge0 (0 = falling, 1 = rising) 
    INTCON2bits.INTEDG1 = 1; // Edge1 (0 = falling, 1 = rising)
    INTCON2bits.INTEDG2 = 0; // Edge2 (0 = falling, 1 = rising)

    // interrupt control 2 register
    if (RCONbits.IPEN) {
        INTCON2bits.TMR0IP = 1; // Overflow interrupt priority bit (0=low priority, 1=high priority)
        INTCON2bits.RBIP = 1;   // Overflow interrupt priority bit (0=low priority, 1=high priority)
    }

    // timer0 control register
    T0CONbits.T0PS   = 0b111; // prescale value (0 = 2, 111 = 256)
    T0CONbits.PSA    = 0;     // prescale (0 = on, 1 = off)
    T0CONbits.T0SE   = 0;     // edge select (1 = high-to-low, 0 = low-to-high)
    T0CONbits.T0CS   = 0;     // clock source (1 = T0CKI, 0 = internal)
    T0CONbits.T08BIT = 0;     // bit (1 = 8-bit, 0 = 16-bit)
    T0CONbits.TMR0ON = 1;     // enable tmr0 (1 = on, 0 = off)

    if (T0CONbits.T08BIT) {
        TMR0H = 0; // Ensure high bytes of timer register are 0 if using 8-bit timer
    }

    // timer1 configuration
    // while there are tmr1, tmr2 and tmr3 clocks, there doesn't 
    // seem to be any tmr1if or tmr1ie for interrupts
    T1CONbits.RD16    = 1; // read write op (1 = one 16-bit, 0 = two 8-bit)
    T1CONbits.T1RUN   = 1; // clock source (1 = Timer1 osc, 0 = other)
    T1CONbits.T1CKPS  = 3; // prescale value (0 = 0 , 11 = 8)
    T1CONbits.T1OSCEN = 1; // oscilator enable (1 = on, 0 = off)
    T1CONbits.T1SYNC  = 1; // sync clock? (1 = 
    T1CONbits.TMR1CS  = 0; // clock source (1 = external clock from RC0/T1OSO/T13CKI pin, 0 = internal clock FOSC/4)
    T1CONbits.TMR1ON  = 0; // tmr1 enable bit (1 = on, 0 = off)

    if (T0CONbits.T08BIT) {
        TMR0H = 0; // Ensure high bytes of timer register are 0 if using 8-bit timer
    }
    TMR0H = 0x4c; // count to (20e6/4/256)
    TMR0L = 0x4b;
    
    // PWM setup
    TRISCbits.TRISC2 = OUTPUT; // PWM RC2, pin 17
    CCP1 = PORTCbits.RC2; // PWM output pin
    T2CON = 0b01111111;
    //T2CON = 0b00000111;
    CCP1CON = 0b00001100; // PWM control register
    PR2 = 0xff; // set period to be 255
    
    // onboard LED setup
    LATDbits.LATD3 = OUTPUT; // LED
    
    // rotary encoder setup
    // RB4 = onboard button
    // RB5 = enc button
    // RB6 = enc pin A
    // RB7 = enc pin B
    LATB  = 0b11000000;
    TRISB = 0b11110000;
   
    // when RB6,RB7 are set as output, latch function works well, pins are held high
    // however, PORTB change-interrupt only occurs on input pins!
    // when RB6,RB7 are set as input, pins don't latch but instead just read input, which is too low to indicate change.
    // thus external pull-ups are needed.
}

int
main(void)
{
    setup();
    LED = OFF;
    val = 0;
   
    enc_dir = 0;
    enc_pos = 0;
    enc_prev_pos = 0;

    // initial read of encoder pins
    enc_prev_pos = (ENC_A << 1) | ENC_B;
     
    while(1) {
	CCPR1L = DC_HI(val);
        CCP1CONbits.DC1B = DC_LO(val);
        delay_ms(10);
    }
}
